#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <clock.h>
#include "emu.h"
#include "utils.h"
#include "dip.h"
#include "fm.h"
#include "psg.h"

WINDOW *emu_window;
GAME *current_game;
MACHINE *machine;
BITMAP *sprites;
int dirty_map[32][16];
int gametic;
int audio_initialised=FALSE, sample_rate, audio_buffer_size;
volatile int timertic;
int use_vsync, frameskip;
int emu_screen_drawn, emu_updated;

int key_left, key_right, key_up, key_down, key_b1, key_b2;
int key_start1, key_start2, key_coin;

GADGET gui_emu_dialog[] = {
   /* (dialog proc)       (x)    (y)   (w)       (h)       (flags)  (d1) (d2)  (dp) */
   { gui_window_proc,     0,     0,    NC_W+256, NC_H+224, 0,       0,   0,    "Emulation" },
   { gui_emu_screen_proc, NC_X,  NC_Y, 256,      224,      0,       0,   0,    NULL },
   { NULL }
};

void throttle_timer(void)
{
   timertic++;
}
END_OF_FUNCTION(throttle_timer);

int Emu_Init(GAME *game)
{
   GUI_PrintLog(FALSE, "Initialising '%s'...", game->long_name);

   /* set everything to zero for safety */
   sprites = NULL;
   gametic = timertic = 0;

   switch(frameskip)
   {
   case 0: GUI_PrintLog(FALSE, "Using automatic frame-skipping to regulate speed."); break;
   case 1: GUI_PrintLog(FALSE, "No speed limiting, drawing every frame."); break;
   case 2: GUI_PrintLog(FALSE, "No speed limiting, drawing every 2nd frame."); break;
   case 3: GUI_PrintLog(FALSE, "No speed limiting, drawing every 3rd frame."); break;
   default: GUI_PrintLog(FALSE, "No speed limiting, drawing every %ith frame.", frameskip); break;
   }
   if (use_vsync)
      GUI_PrintLog(FALSE, "Synchronising to vertical retrace before screen update.");

   LOCK_VARIABLE(timertic);
   LOCK_FUNCTION(throttle_timer);
   install_int_ex(&throttle_timer, BPS_TO_TIMER(game->machine->interrupt_rate));

   for (current_cpu=0;current_cpu<game->machine->num_cpus;current_cpu++)
   {
      game->machine->cpu[current_cpu].iperiod =
         ( game->machine->cpu[current_cpu].hz
         / game->machine->interrupt_rate )
         / game->machine->interleave;
      game->machine->cpu[current_cpu].image = NULL;

      game->machine->cpu[current_cpu].regs = NULL;
   }

   current_game = game;
   machine = game->machine;
   if (!Emu_LoadROMList()) return FALSE;
   if (!Emu_InitAudioSystem()) return FALSE;
   if (!machine->Init()) return FALSE;

   for (current_cpu=0;current_cpu<game->machine->num_cpus;current_cpu++)
   {
      game->machine->cpu[current_cpu].regs = (byte *)malloc(
         z80_get_context_size() * sizeof(byte));
      if (!game->machine->cpu[current_cpu].regs) return FALSE;
   }

   MemDump_Update(NULL, NULL);
   if (emu_window)
   {
      if (use_vsync) emu_window->flags |= W_VSYNC;
      else emu_window->flags &= ~W_VSYNC;
   }
   emu_screen_drawn = emu_palette_dirty = emu_updated = FALSE;

   DIP_LoadSwitches();

   Emu_Reset(NULL, NULL);
   GUI_PrintLog(FALSE, "Emulation OK.");
   return TRUE;
}

int Emu_LoadROMList(void)
{
   ROM_LIST *list;
   char basename[300];

   GUI_PrintLog(TRUE, "Loading ROMs...");
   GUI_SetMouseCursor(gui_data[GUI_BUSY].dat, 0, 0);

   /* zero the list to be safe */
   for (list=current_game->rom_list;list->filename;list++)
   {
      list->rom = NULL;
   }

   /* now load them */
   for (list=current_game->rom_list;list->filename;list++)
   {
      /* construct directory name */
      strcpy(basename, current_game->short_name);
      strcat(basename, "/");
      strcat(basename, list->filename);

      list->rom = Emu_LoadRom(basename, list->size, list->checksum);
      if (!list->rom)
      {
         GUI_SetMouseCursor(gui_data[GUI_CURSOR].dat, 1, 1);
         return FALSE;
      }
   }

   GUI_PrintLog(FALSE, "ROMs loaded OK.");
   GUI_SetMouseCursor(gui_data[GUI_CURSOR].dat, 1, 1);
   return TRUE;
}

void Emu_FreeROMList(void)
{
   ROM_LIST *list;
   GUI_PrintLog(TRUE, "Releasing ROMs.");

   for (list=current_game->rom_list;list->filename;list++)
   {
      if (list->rom) free(list->rom);
   }
}

/* This will blit from src to dest (like Allergo), but
   can also flip horizontally and convert through the colour map.
   It does not do transparency or clipping */
void Emu_Blit(BITMAP *src, BITMAP *dest, int sx, int sy,
                                         int dx, int dy,
                                         int x_flip, int colourbase)
{
   int tw,th;
   byte *psrc, *pdest, c;
   int *map;

   map = &palette_map[colourbase];

   th = 16;
   if (x_flip)
   {
      do {
         tw = 16;
         psrc = &(src->line[sy][sx+16-1]);
         pdest = &(dest->line[dy][dx]);

         do {
            c = *psrc--;
            *pdest++ = map[c];
         } while(--tw);
      
         sy++;
         dy++;
      } while(--th);
   } else {
      do {
         tw = 16;
         psrc = &(src->line[sy][sx]);
         pdest = &(dest->line[dy][dx]);

         do {
            c = *psrc++;
            *pdest++ = map[c];
         } while(--tw);
      
         sy++;
         dy++;
      } while(--th);
   }
}

/* This will blit from src to dest, with clipping, flipping, colouring, etc */
void Emu_MaskedBlit(BITMAP *src, BITMAP *dest, int sx, int sy,
                                               int dx, int dy,
                                               int x_flip, int y_flip,
                                               int colourbase)
{
   int lgap, tgap, w, tw, th;
   byte *psrc, *pdest, c;
   int *map;

   w = th = 16;
   if (dx <= -w || dx >= dest->w) return;
   if (dy <= -th || dy >= dest->h) return;

   lgap = tgap = 0;
   if (x_flip) sx += w-1;
   if (y_flip) sy += th-1;

   if (dx < 0)
   {
      lgap -= dx;
      w += dx;
      dx = 0;
   }
   if (dy < 0)
   {
      tgap -= dy;
      th += dy;
      dy = 0;
   }
   if (dx+w > dest->w)
   {
      w = dest->w - dx;
   }
   if (dy+th > dest->h)
   {
      th = dest->h - dy;
   }

   map = &palette_map[colourbase];
   if (x_flip && y_flip) /* flipped both ways */
   {
      sx -= lgap;
      sy -= tgap;
      do {
         tw = w;
         psrc = &(src->line[sy][sx]);
         pdest = &(dest->line[dy][dx]);

         do {
            c = *psrc--;
            if (c) *pdest = map[c];
            pdest++;
         } while(--tw);

         sy--;
         dy++;
      } while(--th);
   } else if (x_flip)   /* horizontally flipped */
   {
      sx -= lgap;
      sy += tgap;
      do {
         tw = w;
         psrc = &(src->line[sy][sx]);
         pdest = &(dest->line[dy][dx]);

         do {
            c = *psrc--;
            if (c) *pdest = map[c];
            pdest++;
         } while(--tw);

         sy++;
         dy++;
      } while(--th);
   } else if (y_flip)   /* vertically flipped */
   {
      sx += lgap;
      sy -= tgap;
      do {
         tw = w;
         psrc = &(src->line[sy][sx]);
         pdest = &(dest->line[dy][dx]);

         do {
            c = *psrc++;
            if (c) *pdest = map[c];
            pdest++;
         } while(--tw);

         sy--;
         dy++;
      } while(--th);
   } else {    /* not flipped */
      sx += lgap;
      sy += tgap;
      do {
         tw = w;
         psrc = &(src->line[sy][sx]);
         pdest = &(dest->line[dy][dx]);

         do {
            c = *psrc++;
            if (c) *pdest = map[c];
            pdest++;
         } while(--tw);

         sy++;
         dy++;
      } while(--th);
   }
}

void Emu_CheckKeys(BITMAP *scr)
{
   MSG newmsg;
   int x,y;
   char *txt, filename[13];

   /* check for pause */
   if (key[KEY_P])
   {
      text_mode(-1);
      txt = "Paused";
      x = (SCREEN_W - text_length(font, txt)) / 2;
      y = (SCREEN_H - text_height(font)) / 2;
      textprintf(screen, font, x-1, y, GUI_BLACK, txt);
      textprintf(screen, font, x+1, y, GUI_BLACK, txt);
      textprintf(screen, font, x, y-1, GUI_BLACK, txt);
      textprintf(screen, font, x, y+1, GUI_BLACK, txt);
      textprintf(screen, font, x, y, GUI_WHITE, txt);

      while(key[KEY_P]);
      while(!key[KEY_P] && !key[KEY_ESC]);
      while(key[KEY_P]);

      newmsg.type = MSG_REFRESH;
      GUI_PostMessage(&global_msg_queue, &newmsg);
   }

   /* check for F12 - screengrab */
   if (key[KEY_F12])
   {
      while(key[KEY_F12]);

      x = 1;
      sprintf(filename, "tnzs%.4i.pcx", x);
      while(!access(filename, F_OK) && x < 9999)
      {
         x++;
         sprintf(filename, "tnzs%.4i.pcx", x);
      }

      if (x >= 9999)
      {
         GUI_PrintLog(TRUE, "Too many screenshots!");
      } else {
         if (save_bitmap(filename, scr, gui_palette))
            GUI_PrintLog(TRUE, "Error writing screen-grab to file '%s'.", filename);
         else
            GUI_PrintLog(FALSE, "Saved screen-grab to file '%s'.", filename);
      }
   }
}

int gui_emu_screen_proc(MSG *msg, WINDOW *window, GADGET *d)
{
   int frames;

   switch(msg->type)
   {
   case MSG_IDLE:
      if (!current_game || emu_updated) return D_O_K;

      Emu_CheckKeys(d->bmp);
   
      if (frameskip)
         frames = frameskip;
      else
         frames = timertic - gametic;

      if (!frames)
      {
         if (use_vsync) frames = 1;
         else return D_O_K;
      }
      if (frames > MAX_FRAME_SKIP) frames = MAX_FRAME_SKIP;

      do {
         Emu_UpdateAudio();
         if (!Emu_DoFrame()) return D_CLOSE;
      } while(--frames);
      gametic = timertic;

      d->flags |= D_DIRTY;
      emu_screen_drawn = FALSE;
      emu_updated = TRUE;
      return D_O_K;
   case MSG_DRAW:
      if (!current_game)
      {
         rectfill(d->bmp, 0, 0, d->w - 1, d->h - 1, GUI_BG_NORMAL);
      } else {
         Emu_UpdateAudio();
         Clock_Start(CLOCK_RENDER);
         machine->DrawScreen(d->bmp);
         Clock_Stop(CLOCK_RENDER);
         frame_count++;
         emu_screen_drawn = TRUE;
         Emu_UpdateAudio();
      }
      return D_O_K;
   case MSG_REFRESH:
      if (emu_screen_drawn)
      {
         if (emu_palette_dirty)
         {
            gui_palette_dirty = TRUE;
            emu_palette_dirty = FALSE;
         }
         emu_updated = FALSE;
      }
      return D_O_K;
   default: return D_O_K;
   }
}

void Emu_RequestRedraw(void)
{
   int x, y;

   /* force the background to be redrawn */
   for (x=0;x<32;x++)
   {
      for (y=0;y<16;y++)
      {
         dirty_map[x][y] = -1;
      }
   }
}

void Emu_UpdateAudio(void)
{
   int n, num_types;
   AUDIO_CHIP *chip;

   Clock_Start(CLOCK_AUDIO);
   num_types = machine->num_audio_chip_types;
   if (num_types)
   {
      chip = &(machine->audio_chip_types[0]);
      do {
         for (n=0;n<chip->num;n++)
         {
            chip->driver->Emulate(n);
         }
         chip++;
      } while(--num_types);
   }
   Clock_Stop(CLOCK_AUDIO);
}

int Emu_InitAudioSystem(void)
{
   int n;
   AUDIO_CHIP *chip;

   if (audio_initialised) return TRUE;

   sample_rate = get_config_int("Audio_settings", "sample_rate", 22050);
   GUI_PrintLog(FALSE, "Initialising audio system at %iHz...", sample_rate);

   audio_buffer_size = sample_rate / machine->interrupt_rate;
   GUI_PrintLog(FALSE, "Audio buffer size=%i bytes", audio_buffer_size);

   if (machine->num_audio_chip_types)
   {
      chip = &(machine->audio_chip_types[0]);
      GUI_PrintLog(FALSE, "Initialising '%s' driver.", chip->driver->description);
      for (n=0;n<machine->num_audio_chip_types;n++)
      {
         if (!chip->driver->Init(chip->num, chip->hz, sample_rate, audio_buffer_size))
         {
            GUI_PrintLog(TRUE, "Cancelling audio initialisation.");
            do {
               chip->driver->Shutdown();
               chip--;
            } while(n--);
            return FALSE;
         }
      }
   }

   audio_initialised = TRUE;
   return TRUE;
}

void Emu_ShutdownAudioSystem(void)
{
   int n;
   AUDIO_CHIP *chip;
   if (!audio_initialised) return;

   GUI_PrintLog(FALSE, "Shutting down audio system...");
   n = machine->num_audio_chip_types;
   if (n)
   {
      chip = &(machine->audio_chip_types[0]);
      do {
         chip->driver->Shutdown();
         chip++;
      } while(--n);
   }

   set_config_int("Audio_settings", "sample_rate", sample_rate);
   audio_initialised = FALSE;
}

void EmuWindow_Shutdown(WINDOW *window)
{
   set_config_int("Emulation_window", "x_position", window->x);
   set_config_int("Emulation_window", "y_position", window->y);
   Emu_Shutdown();
}

void Emu_Shutdown(void)
{
   GUI_PrintLog(FALSE, "Shutting down emulation...");

   if (!current_game) return;

   DIP_Shutdown(dip_switch_window);
   Emu_ShutdownAudioSystem();
   Emu_FreeROMList();
   machine->Shutdown();
   if (sprites) destroy_bitmap(sprites);

   for (current_cpu=0;current_cpu<machine->num_cpus;current_cpu++)
   {
      if (machine->cpu[current_cpu].image)
         free(machine->cpu[current_cpu].image);

      if (machine->cpu[current_cpu].regs)
         free(machine->cpu[current_cpu].regs);
   }     
   
   if (use_vsync) set_config_string("Options", "use_vsync", "yes");
   else set_config_string("Options", "use_vsync", "no");
   set_config_int("Options", "frameskip", frameskip);

   current_game = NULL;
}

byte *Emu_LoadRom(char *filename, int expected_size, unsigned int expected_checksum)
{   
   FILE *file;
   int size, n;
   word sum, xor, t;
   unsigned int checksum;
   byte *data, t1,t2;

   sum = xor = 0;
   file = fopen(filename, "rb");
   if (!file)
   {
      GUI_PrintLog(TRUE, "Cannot open ROM file '%s'.", filename);
      GUI_MessageBox(MB_OK, NULL, "Cannot open ROM file '%s'.", filename);
      return NULL;
   }
   fseek(file, 0, SEEK_END);
   size = ftell(file);
   fseek(file, 0, SEEK_SET);
   data = (byte *)malloc(size);
   if (!data)
   {
      GUI_PrintLog(TRUE, "Out of memory while loading ROM file '%s'.", filename);
      GUI_MessageBox(MB_OK, NULL, "Out of memory while loading ROM file '%s'.", filename);
      return NULL;
   }
   if (!fread(data, 1, size, file))
   {
      fclose(file);
      GUI_PrintLog(TRUE, "Cannot read from ROM file '%s'.", filename);
      GUI_MessageBox(MB_OK, NULL, "Cannot read from ROM file '%s'.", filename);
      return NULL;
   }
   fclose(file);

   if (size != expected_size)
   {
      GUI_PrintLog(TRUE, "Warning: ROM file '%s' was %iKb when it should be %iKb.",
                     filename, size/1024, expected_size/1024);
      GUI_PrintLog(TRUE, "The game might not function as expected.");
      GUI_MessageBox(MB_OK, "The game might not function as expected.",
                            "Warning: ROM file '%s' was %iKb when it should be %iKb.",
                            filename, size/1024, expected_size/1024);
   } else {
      for (n=0;n<size;n+=2)
      {
         t1 = data[n];
         t2 = data[n+1];
         t = (t1 << 8) + t2;
         sum += t;
         xor ^= t;
      }
      checksum = ((dword)sum << 16) | xor;
      if (checksum == expected_checksum)
      {
         GUI_PrintLog(FALSE, "Loaded ROM file '%s' successfully (%iKb).",
                      filename, expected_size/1024);
         GUI_PrintLog(FALSE, "Checksum verified for '%s' (0x%.8x).",
                      filename, checksum);
      } else {
         GUI_PrintLog(TRUE, "Warning: ROM file '%s' has checksum 0x%.8x when it should be 0x%.8x.",
                      filename, checksum, expected_checksum);
         GUI_PrintLog(TRUE, "The game might not function as expected.");
         GUI_MessageBox(MB_OK, "The game might not function as expected.",
                               "Warning: ROM file '%s' has checksum 0x%.8x when it should be 0x%.8x.",
                               filename, checksum, expected_checksum);
      }
   }
   return data;
}

inline void DrawPlanarByte(BITMAP *bmp, int plane, int x, int y, byte c)
{
   int count, t;
   count = 8;
   while(count)
   {
      t = (c&128) >> 7;
      bmp->line[y][x] |= (t << plane);
      x++;
      c <<= 1;
      count--;
   }
}

void Emu_AddPlane(BITMAP *dest, byte *src, int plane, int tile, int count)
{
   int x, y;
   byte c;

   x = 0;
   y = tile*16;
   do {
      c = *src++;
      DrawPlanarByte(dest, plane, x, y, c);
      y++;
      if (!(y&7))
      {
         if (x == 8)
         {
            x = 0;
         } else {
            x = 8;
            y -= 8;
         }
      }
   } while (--count);
}
